try: from .util.demo import *
except ImportError as e:
  if e.msg.lower() == "no module named 'ion'": e.msg = ("ion module isn't installed, type 'pip install ion-numworks' to install it.",)
  print(f"Error: demo can't be load! ({e.msg})")



from sdl2 import SDL_Rect, SDL_FillRect, SDL_BlitSurface, SDL_BlitScaled, SDL_CreateRGBSurfaceWithFormat, SDL_PIXELFORMAT_RGB888
from sdl2.ext import prepare_color
from ctypes import cast, byref, POINTER, c_uint32, c_uint8
from .vars import Vars

__all__ = ["Draw"]


class Draw:
  new_area = lambda *area: SDL_Rect(*[int(i*Vars.zoom_ratio) for i in (area[0] if len(area)==1 and type(area[0]) in (list, tuple) else area)])
  new_surface = lambda width, height: SDL_CreateRGBSurfaceWithFormat(0, int(width*Vars.zoom_ratio), int(height*Vars.zoom_ratio), 32, SDL_PIXELFORMAT_RGB888).contents
  blit = lambda dest, source, area=None: SDL_BlitSurface(source, None, dest, Draw.new_area(area) if area else None)
  blit_scaled = lambda dest, source, area=None: SDL_BlitScaled(source, None, dest, Draw.new_area(area+(source.w, source.h) if len(area) == 2 else area) if area else None)
  rect = lambda dest, color, area=None: SDL_FillRect(dest, Draw.new_area(area) if area else None, prepare_color(color, dest))
  pixel = lambda dest, color, x, y: SDL_FillRect(dest, Draw.new_area(x, y, 1, 1), prepare_color(color, dest))
  string = lambda dest, font, text, x, y, color=None: Draw.blit_scaled(dest, font.render(text, color=color), (x, y))

  def get_at(source, x, y):
    pixel = cast(byref(cast(source.pixels, POINTER(c_uint8)).contents, source.pitch*Vars.zoom_ratio*y+x*Vars.zoom_ratio*source.format.contents.BytesPerPixel), POINTER(c_uint32)).contents.value
    return (pixel&0xff0000)>>8, (pixel&0x00ff00)>>4, (pixel&0x0000ff)>>0



"""This is a demo code
Source: https://my.numworks.com/python/zetamap/snake_lite
"""
print("""This is a demo code
Source: https://my.numworks.com/python/zetamap/snake_lite""")
__all__ = []

#############################################################

from random import randint as ri
from kandinsky import fill_rect as fr, draw_string as ds
from ion import keydown as kd
from time import monotonic as mt, sleep as sl

DECODE=lambda v: ((v&0xfff000)>>12, v&0xfff)
class Snake:
 c=[(50,50,50),(240,240,240),(0,255,0),"green","red","blue","#7ea2ce"]
 def __init__(s,leader_board=(),speed=10,power=2,size=10,score=1,inf_snake=False,gost=False,darkmode=True,rainbow=False,walls=True):
  if inf_snake: power=float("inf")
  if not darkmode: s.c[0],s.c[1]=s.c[1],s.c[0]
  try: from kandinsky import get_keys
  except ImportError:
   try: from ion import get_keys
   except ImportError: s.c[6]="#ffb531"
   else: s.c[6]="#c53431"
  s.lb,s.s,s.p,s.m,s.ti,s.g,s.go,s.r,s.w,s.bo=leader_board,1/speed,power,score,size,(320//size-1,220//size-1),gost,rainbow,walls,bool(walls)
 def nc(s):
  s.ch=[ri(s.bo,s.g[i]-s.bo) for i in range(2)]
  while s.ch[0]<<12|s.ch[1] in s.b: s.ch=[ri(s.bo,s.g[i]-s.bo) for i in range(2)]
  fr(s.ch[0]*s.ti,2+s.ch[1]*s.ti,s.ti,s.ti,s.c[4])
 def dl(s):
  fr(0,2,320,220,"black")
  fr(0,2,(s.g[0]+1)*s.ti,(s.g[1]+1)*s.ti,s.c[5 if s.w else 0])
  if s.w: fr(s.ti//2,2+s.ti//2,s.g[0]*s.ti,s.g[1]*s.ti,s.c[0])
 def dt(s,text,x,y):
  for i,l in enumerate(text.replace('\t', "    ")):
   ds(l,x+10*i,y,s.c[1],s.c[0])
   if l!=' ': sl(0.02)
 def ub(s,l1,l2):
  fr(20,82,280,50,s.c[4])
  fr(25,87,270,40,s.c[0])
  s.dt(l1,25,87)
  s.dt(l2,25,107)
  sl(0.2)
  while 1:
   sl(0.01)
   if kd(4): return 1
   elif kd(17): return 0
 def wu(s):
  s.t,s.ta,s.te=mt(),(s.b[-1], DECODE(s.b[-1])),DECODE(s.b[0])
  #s.d=[]
  if kd(0) and s.d[0]!=1: s.d=[-1,0]
  elif kd(1) and s.d[1]!=1: s.d=[0,-1]
  elif kd(2) and s.d[1]!=-1: s.d=[0,1]
  elif kd(3) and s.d[0]!=-1: s.d=[1,0]
  fr(s.te[0]*s.ti,2+s.te[1]*s.ti,s.ti,s.ti,[ri(0,255) for i in range(3)] if s.r else s.c[3])
  for x in range(len(s.b)-1,0,-1): s.b[x]=s.b[x-1]
  s.b[0]=((s.te[0]+s.d[0])%s.g[0])<<12|((s.te[1]+s.d[1])%s.g[1])
  if s.a>0:
   s.b.append(s.ta[0])
   s.a-=1
  elif s.ta[0] not in s.b: fr(s.ta[1][0]*s.ti,2+s.ta[1][1]*s.ti,s.ti,s.ti,s.c[0])
  if s.b[0]==s.ch[0]<<12|s.ch[1]:
   s.a,s.sc=s.a+s.p,s.sc+s.m
   s.nc()
  s.te=DECODE(s.b[0])
  if not s.go:
   if s.w and (s.te[0]==0 or s.te[1]==0 or s.te[0]==s.g[0] or s.te[1]==s.g[1]): return 0
   for x in range(1,len(s.b)):
    if s.b[x]==s.b[0]: return 0
  fr(s.te[0]*s.ti,2+s.te[1]*s.ti,s.ti,s.ti,s.c[2])
  ds(str(s.sc),7,7,s.c[1],s.c[0])
  return 1
 def start(s):
  fr(0,0,320,2,s.c[6])
  try:
   while 1:
    s.b,s.a,s.d,s.sc=[(s.g[0]//2)<<12|(s.g[1]//2)],4,(1,0),0
    s.dl()
    s.nc()
    while s.wu():
     if kd(17):
      s.ub("\t\tGame Paused!","  (OK, DELETE = Continue)")
      sl(0.2)
      s.dl()
      fr(s.ch[0]*s.ti,2+s.ch[1]*s.ti,s.ti,s.ti,s.c[4])
      for i in s.b:
       s.te=DECODE(i)
       fr(s.te[0]*s.ti,2+s.te[1]*s.ti,s.ti,s.ti,[ri(0,255) for i in range(3)] if s.r else s.c[3])
     while mt()-s.t<s.s: sl(0.01)
    if not s.ub(' '*(3-len(str(s.sc))//2)+"You lose!\tScore: "+str(s.sc),"(OK = Retry, DELETE = Quit)"): break
   s.te=0
   print("Your score:",s.sc,"\nLeader board: (default settings)")
   for i,sc in enumerate(s.lb):
    if len(sc[0])>s.te:s.te=len(sc[0])
    print(' %s- %s: %s'%(i+1,sc[0],sc[1]))
   fr(0,2,320,220,s.c[0])
   s.dt("Your score: "+str(s.sc),5,7)
   s.dt("Leader board:",5,32)
   for i,sc in enumerate(s.lb): s.dt(sc[0]+':'+' '*(s.te-len(sc[0])+2)+str(sc[1]),60,52+20*i)
   l=['>'*i+' '*(4-i)+s.lb[0][0]+':'+' '*(s.te-len(s.lb[0][0])+2)+str(s.lb[0][1])+' '*(4-i)+'<'*i for i in range(4)]
   while not kd(4) and not kd(17):
    for i in l:
     sl(0.2)
     if kd(4) or kd(17): break
     ds(i,20,52,s.c[6],s.c[0])
  except KeyboardInterrupt: pass

Snake(( # Leader board: ("name", score),
 ("Alteur",38),
 ("ZetaMap",30),
 ("Nios",23),
 ("Alpha6Frost",17),

), # Game settings: power, speed, size, score, inf_snake, darkmode, rainbow, gost, walls
#  walls=False,
#  size=5, # 1 -> 80
#  darkmode=False
).start()



from tkinter import Tk, Frame, Menu, IntVar

from tkinter.dialog import Dialog
from tkinter.filedialog import asksaveasfilename
from tkinter.font import Font

from sdl2 import SDL_CreateWindowFrom, SDL_DestroyWindow, SDL_FreeSurface#, SDL_WINDOW_OPENGL
from sdl2.ext import Window, load_img, init

from os import getcwd, system
from webbrowser import open as open_link, register_standard_browsers  # To open links in help menu
register_standard_browsers() # register now to avoid a long wait when opening a link
del register_standard_browsers
from datetime import datetime

# Internal
from .vars import *
from .draw import Draw
from .color import Colors
from . import limiter

__all__ = ["Gui"]


class Gui:
  tkmaster:Tk = None
  paused = False
  already_paused = False
  drawable = None
  _main_thread_pid = None # because thread.native_pid don't exist in python<3.8, so use the other way
  heap_set = False

  def created():
    if not Gui.tkmaster: raise RuntimeError("Gui: an instance must be created")

  def _unpause_after_wrapper(method):
    def _wrap(*a, **k):
      method(*a, **k)
      if not Gui.already_paused: Gui.paused = False

    _wrap.__name__ = method.__name__
    return _wrap

  # Patched menu buttons to unpause script when is clicked
  _add_command = lambda menu, **kwargs: menu.add_command(**dict([(k, Gui._unpause_after_wrapper(v)) if "command" in k else (k, v) for k, v in kwargs.items()]))
  _add_radiobutton = lambda menu, **kwargs: menu.add_radiobutton(**dict([(k, Gui._unpause_after_wrapper(v)) if "command" in k else (k, v) for k, v in kwargs.items()]))

  def __init__(_, tkmaster):
    # This is now, the video subsystem of SDL will be initialized
    init(video=True)

    Gui.tkmaster = tkmaster
    if Vars.is_linux:
      Gui.linux_menu_font = Font(tkmaster, family="SegoeUI", size=9)
      Gui.tkmaster.option_add("*font", Gui.linux_menu_font)
    Gui.data = StateData()

    # Create frame area, sdl2 windows and pack everything
    Gui.head_frame = Frame(tkmaster)
    Gui.screen_frame = Frame(tkmaster)
    Gui.head_frame.pack()
    Gui.screen_frame.pack()
    Gui.tkmaster.update()
    Gui.head = Window('',(0,0))
    Gui.screen = Window('',(0,0))

    Gui.menu = Menu(tkmaster)
    # Menus
    ## Help menu
    about = Menu(tearoff=False)
    Gui._add_command(about, label="GitHub project",     command=lambda: open_link("https://github.com/ZetaMap/Kandinsky-Numworks"))
    Gui._add_command(about, label="Documentation",      command=lambda: open_link("https://github.com/ZetaMap/Kandinsky-Numworks#usable-content"))
    Gui._add_command(about, label="An issue? Open one", command=lambda: open_link("https://github.com/ZetaMap/Kandinsky-Numworks/issues/new"))
    Gui._secret_var = 0
    def secret():
      Gui._secret_var += 1
      open_link("https://github.com/ZetaMap")
      if Gui._secret_var == 3:
        about.add_separator()
        the_secret = lambda: system(f"start pythonw {Vars.path}../3d_demo.py") if Vars.is_windows else system(f"python3 {Vars.path}../3d_demo.py &")
        Gui._add_command(about, label="3D engine demo", command=the_secret)
    Gui._add_command(about, label="Made by ZetaMap",    command=secret)
    Gui.menu.add_cascade(label="About", menu=about)

    ## Help menu
    help = Menu(tearoff=False)
    shortcuts = Menu(tearoff=False)

    Gui._add_command(shortcuts, label="CTRL+Q: Close window",    state="disabled", activebackground="#F0F0F0")
    Gui._add_command(shortcuts, label="CTRL+O: change OS",       state="disabled", activebackground="#F0F0F0")
    Gui._add_command(shortcuts, label="CTRL+M: change Model",    state="disabled", activebackground="#F0F0F0")
    Gui._add_command(shortcuts, label="CTRL+S: take screenshot", state="disabled", activebackground="#F0F0F0")
    Gui._add_command(shortcuts, label="CTRL+P: pause/resume",    state="disabled", activebackground="#F0F0F0")
    Gui._add_command(shortcuts, label="CTRL+Z: change zoom",     state="disabled", activebackground="#F0F0F0")
    help.add_cascade(label="Shortcuts", menu=shortcuts)

    help.add_separator()
    Gui._add_command(help, label="Ion keyboard", command=lambda: open_link("https://github.com/ZetaMap/Ion-Numworks#associated-keyboard-keys"))
    Gui._add_command(help, label="Check FAQ",    command=lambda: open_link("https://github.com/ZetaMap/Kandinsky-Numworks/blob/main/FAQ.md"))
    Gui.menu.add_cascade(label="Help", menu=help)

    ## Options menu
    Gui.options = Menu(tearoff=False)

    ## OS menu
    Gui.os_mode = IntVar(value=Vars.selected_os)
    new = Menu(tearoff=False)
    for i, mode in enumerate(Config.os_list):
      Gui._add_radiobutton(new, label=mode["name"], variable=Gui.os_mode, value=i, command=Gui.update_data)
      mode["battery"] = load_img(Vars.path+mode["battery"])
    new.add_separator()
    Gui._add_command(new, label="Initial mode: "+Config.os_list[Vars.selected_os]["name"],
                          state="disabled", activebackground="#F0F0F0")
    Gui.options.add_cascade(accelerator="CTRL+O", label="OS", menu=new)

    ## Model menu
    Gui.model_mode = IntVar(value=Vars.selected_model)
    new = Menu(tearoff=False)
    for i, mode in enumerate(Config.model_list):
      Gui._add_radiobutton(new, label=mode["name"], variable=Gui.model_mode, value=i, command=Gui.update_data,
        **({"state": "disabled", "activebackground": "#F0F0F0"} if mode.get("disabled", False) else {}))
    Gui.options.add_cascade(accelerator="CTRL+M", label="Model", menu=new)
    new.add_separator()
    Gui._add_command(new, label="Initial model: "+Config.model_list[Vars.selected_model]["name"],
                          state="disabled", activebackground="#F0F0F0")

    Gui.zoom = IntVar(value=Vars.zoom_ratio)
    Gui.increase_font = IntVar()
    Gui.zoom.trace_add("write", Gui.set_zoom)
    Gui.increase_font.trace_add("write", Gui.set_zoom)
    new = Menu(tearoff=False)
    for i in range(Config.zoom_max): Gui._add_radiobutton(new, label=f"x{i+1}", variable=Gui.zoom, value=i+1)
    if Vars.is_linux: new.add_checkbutton(label="Increase font size", variable=Gui.increase_font)
    Gui.options.add_cascade(accelerator="CTRL+Z", label="Zoom", menu=new)

    Gui.options.add_separator()

    ## Screenshot button
    Gui.options.add_command(accelerator="CTRL+S", label="Screenshot", command=Gui.askscreenshot)

    ## Pause button
    def state(states=["Pause", "Resume"]):
      Gui.already_paused = not Gui.already_paused
      Gui.paused = Gui.already_paused
      Gui.options.entryconfig(states[not Gui.already_paused], label=states[Gui.already_paused])
    Gui.options.add_command(accelerator="CTRL+P", label="Pause", command=state)

    Gui.menu.add_cascade(label="Options", menu=Gui.options)

  def set_zoom(*_, force=False):
    Gui.created()

    Gui.paused = True # Give time to change var, destroy and rebuild window
    last_zoom = Vars.zoom_ratio
    Vars.zoom_ratio = Gui.zoom.get()
    if Vars.is_linux: Gui.linux_menu_font.config(size=9*(Vars.zoom_ratio if Gui.increase_font.get() else 1))

    if not force and last_zoom == Vars.zoom_ratio:
      # No need to continue
      if not Gui.already_paused: Gui.paused = False
      return

    SDL_DestroyWindow(Gui.head.window)
    SDL_DestroyWindow(Gui.screen.window)
    Gui.head_frame.config(width=Vars.screen[0]*Vars.zoom_ratio, height=Vars.head_size*Vars.zoom_ratio)
    Gui.screen_frame.config(width=Vars.screen[0]*Vars.zoom_ratio, height=Vars.screen[1]*Vars.zoom_ratio)
    Gui.tkmaster.update()
    Gui.head.window = SDL_CreateWindowFrom(Gui.get_widget_id(Gui.head_frame))
    Gui.screen.window = SDL_CreateWindowFrom(Gui.get_widget_id(Gui.screen_frame))
    Gui.screen_surf = Gui.screen.get_surface()

    last_drawable = Gui.drawable
    Gui.drawable = Draw.new_surface(*Vars.screen)
    Draw.rect(Gui.drawable, Colors.fix2(Colors.white))
    Draw.blit_scaled(Gui.drawable, last_drawable)
    Gui.update_data()

    Gui.center_window()
    if not Gui.already_paused: Gui.paused = False

  def get_widget_id(widget):
    """Method to get id of widget, will be replaced by mac_patcher.py if needed"""
    return widget.winfo_id()

  def destroy():
    Gui.created()

    SDL_FreeSurface(Gui.drawable)
    SDL_FreeSurface(Gui.head_surface)
    Gui.head.close()
    Gui.screen.close()
    Gui.tkmaster.destroy()
    Gui.tkmaster.quit()
    Gui.tkmaster = None

  def update_value(int_var, values, min=0):
    Gui.created()
    v = int_var.get()+1
    if v > len(values)-1: v = min

    if not values[v].get("disabled", False):
      int_var.set(v)
      Gui.update_data()

  def update_data():
    Gui.created()
    if Gui.head.window is None: return

    Gui.data(**Config.os_list[Gui.os_mode.get()], model=Config.model_list[Gui.model_mode.get()]["ratio"])
    Gui.head_surface = Draw.new_surface(Vars.screen[0], Vars.head_size)

    Draw.rect(Gui.head_surface, Gui.data.color)
    Draw.string(Gui.head_surface, Config.small_font, Gui.data.unit, 5, 0)
    Draw.string(Gui.head_surface, Config.small_font, "PYTHON", Vars.screen[0]//2-21, 1)
    x = Vars.screen[0]-20
    if Gui.data.clock:
      Draw.string(Gui.head_surface, Config.small_font, datetime.now().strftime("%H:%M"), x-20, 1)
      x -= 40
    Draw.blit_scaled(Gui.head_surface, Gui.data.battery, (x, 4))
    # try to blit it now on screen
    try: Draw.blit(Gui.head.get_surface(), Gui.head_surface)
    except: pass

    # Set new heap of python
    if Gui._main_thread_pid:
      try: limiter.set_heap(Gui._main_thread_pid, Gui.data.heap)
      except (AssertionError, OSError): Gui.heap_set = False
      else: Gui.heap_set = True

  def config(create_gui=False):
    Gui.created()

    # Config main window
    Gui.tkmaster.title(Vars.app_name)
    if Vars.is_windows: Gui.tkmaster.iconbitmap(default=Vars.path+"app.ico")
    else:
      from tkinter import PhotoImage
      Gui.tkmaster.iconphoto(True, PhotoImage(file=Vars.path+"app.png"))
    Gui.tkmaster.resizable(False, False)
    if create_gui: Gui.tkmaster.config(menu=Gui.menu)
    Gui.tkmaster.eval('tk::PlaceWindow . center') # I must keep this otherwise the SDL windows do not configure correctly
    Gui.center_window()

    # Bind shorcuts
    Gui.tkmaster.bind("<Control-o>", lambda _: Gui.update_value(Gui.os_mode, Config.os_list))
    Gui.tkmaster.bind("<Control-m>", lambda _: Gui.update_value(Gui.model_mode, Config.model_list))
    Gui.tkmaster.bind("<Control-s>", lambda _: Gui.askscreenshot())
    def state(*_, states=["Pause", "Resume"]):
      Gui.already_paused = not Gui.already_paused
      Gui.paused = Gui.already_paused
      Gui.options.entryconfig(states[not Gui.already_paused], label=states[Gui.already_paused])
    Gui.tkmaster.bind("<Control-p>", state)
    Gui.tkmaster.bind("<Control-z>", lambda _: Gui.update_value(Gui.zoom, [{}]*(Config.zoom_max+1), 1))

    ### Windows patchs
    if Vars.is_windows:
      # Bind events to pause, if dragged or user interact with gui
      def drag_event(e):
        if e.widget is Gui.tkmaster:
          if Gui._drag_window_event_id:
            Gui.tkmaster.after_cancel(Gui._drag_window_event_id)
            Gui.paused = True
          Gui._drag_window_event_id = Gui.tkmaster.after(200, drag_event_stop)
      def drag_event_stop():
        Gui._drag_window_event_id = ''
        if not Gui.already_paused: Gui.paused = False
      drag_event_stop()
      Gui.tkmaster.bind("<Configure>", drag_event)

      # Pause on opening menu and unpause when click on frames
      def pause(*_): Gui.paused = True
      unpause = Gui._unpause_after_wrapper(lambda *_: None)

      Gui.menu["postcommand"] = pause
      Gui.head_frame.bind("<Button-1>", unpause)
      Gui.screen_frame.bind("<Button-1>", unpause)
    ###

    Gui.set_zoom(force=True) # Force set zoom of window
    Gui.update_data() # Set default data
    Gui.refresh() # Refresh SDL windows and Tkinter frames

  def center_window():
    Gui.created()
    Gui.tkmaster.geometry(f"+{max(0, Gui.tkmaster.winfo_screenwidth() //2-Gui.tkmaster.winfo_width() //2)}"
                          f"+{max(0, Gui.tkmaster.winfo_screenheight()//2-Gui.tkmaster.winfo_height()//2)}")

  def refresh():
    Gui.created()

    Gui.head.refresh()
    Draw.rect(Gui.screen_surf, Colors.black)
    Draw.blit(Gui.screen_surf, Gui.drawable)
    Gui.screen.refresh()
    Gui.tkmaster.update_idletasks()
    Gui.tkmaster.update()

  def askscreenshot():
    Gui.created()
    Gui.paused = True # Pause events
    file = (getcwd().replace("\\\\", "\\"), datetime.now().strftime("Screenshot_%Y%m%d-%H%M%S.png"))

    # Create new surface to blit head and screen into
    surf = Draw.new_surface(Vars.screen[0], Vars.head_size+Vars.screen[1])
    Draw.blit(surf, Gui.head_surface)
    Draw.blit(surf, Gui.drawable, (0, Vars.head_size))

    try: error = Config.save_image(surf, file[1]) < 0
    except: error = 1

    conf = {"title": "Screenshot", "text": "Screenshot saved at: \n"+('\\' if Vars.is_windows else '/').join(file), "bitmap": "info", "default": 1, "strings": ("Open folder", "OK")}
    if error: conf.update({"text": "Error, can't write in folder: \n"+file[0], "bitmap": "error", "strings": ("Save as", "OK")})

    if not Dialog(Gui.tkmaster, conf).num:
      if error:
        file = asksaveasfilename(defaultextension="png", filetypes=Vars.image_formats, initialfile=file[1], title="Save screenshot at")
        if file != '':
          try: Config.save_image(surf, file)
          except: pass
      else: system(f"{'explorer' if Vars.is_windows else 'open' if Vars.is_macos else 'xdg-open'} \"{file[0]}\" 2> {'nul' if Vars.is_windows else '/dev/null'}")

    SDL_FreeSurface(surf) # destroy screenshot surface
    if not Gui.already_paused: Gui.paused = False # Screenshot finished, unpause events

  def askscriptend():
    return not Dialog(Gui.tkmaster, {
      "title": "Script end", "text": "Script finished! \nClose window?".ljust(50),
      "bitmap": "question", "default": 0, "strings": ("Yes", "No")
    }).num



from .tests import Tests
from .vars import Vars

class Colors:
  """https://github.com/numworks/epsilon/blob/master/kandinsky/include/kandinsky/color.h"""

  blue    = (0,   0,   248)
  b       = blue
  red     = (248, 0,   0)
  r       = red
  green   = (80,  192, 0)
  g       = green
  yellow  = (248, 252, 0)
  y       = yellow
  brown   = (136, 112, 80)
  black   = (0,   0,   0)
  k       = black
  white   = (248, 252, 248)
  w       = white
  pink    = (248, 168, 176)
  orange  = (248, 132, 24)
  purple  = (104, 44,  120)
  grey    = (160, 164, 160)
  gray    = grey
  cyan    = (0,   252, 248)
  magenta = (248, 4,   136)
  COLORS = {k: v for k, v in locals().items() if not k.startswith('_')}

  fix = lambda r, g, b, expand=Vars.selected_os <= 1: (
    (Colors.expand((r>>3)&0x1f, 5) if expand else ((r>>3)&0x1f)<<3),
    (Colors.expand((g>>2)&0x3f, 6) if expand else ((g>>2)&0x3f)<<2),
    (Colors.expand((b>>3)&0x1f, 5) if expand else ((b>>3)&0x1f)<<3))
  fix2 = lambda *values, expand=Vars.selected_os <= 1: (
    tuple(Colors.expand(Colors.contract(c, i%2, 3), 5+i%2) for i, c in enumerate(*values)) 
      if expand else
    tuple(Colors.contract(c, i%2, 3)<<(3-i%2) for i, c in enumerate(*values)))
  contract = lambda v, nBits, length=8: (v>>(length-nBits))&(2**(4+length-nBits)-1)
  expand = lambda v, nBits, length=8: (v<<(length-nBits))|(v>>(nBits-(length-nBits)))

  def convert(rgbOrName):
    _type = type(rgbOrName)

    if _type == str:
      if rgbOrName in Colors.COLORS: return Colors.fix(*Colors.COLORS[rgbOrName])
      elif rgbOrName.startswith('#'):
        if len(rgbOrName) != 7: raise ValueError("RGB hex values are 6 bytes long")

        try: return Colors.fix(*[int(rgbOrName[i:i+2], 16) for i in range(1, len(rgbOrName), 2)])
        except ValueError as e:
          e.args = (f"invalid literal for int() with base 16: '{rgbOrName[1:]}'",)
          raise
      else:
        try: ratio = float(rgbOrName)
        except ValueError as e:
          e.args = ("invalid syntax for number",)
          raise
        else:
          if 0 <= ratio <= 1: return tuple([int(255*ratio) for _ in range(3)])
          else: raise ValueError("Gray levels are between 0.0 and 1.0")

    elif _type == int: raise ValueError("Int are not colors")

    Tests.list(rgbOrName)
    if len(rgbOrName) != 3: raise ValueError("Color needs 3 components")
    return Colors.fix(*[int(c) for c in rgbOrName if type(c) == float or Tests.int(c)])



from sdl2 import SDL_RWFromFile, SDL_SaveBMP_RW
from sdl2.sdlimage import IMG_SavePNG_RW, IMG_SaveJPG_RW
from sdl2.ext.ttf import FontManager

import sys, os

__all__ = [
  "Vars",
  "Config",
  "StateData"
]


class Vars:
  base_name = __module__.split('.')[1]
  path = __file__[:__file__.rindex(base_name)+len(base_name)+1]+"data/"
  # Test if path is correct
  if not os.path.exists(path): path = __file__[:__file__.rindex(base_name)]+"data/"

  is_windows = sys.platform.startswith("win")
  is_macos = sys.platform.startswith("darwin")
  is_linux = sys.platform.startswith("linux") or (not is_windows and not is_macos)
  app_name = "Kandinsky Emulator"
  head_size = 18
  screen = (320, 222)
  zoom_ratio = selected_os = selected_model = 1

  image_formats = [("PNG", ".png"), ("Bitmap", ".bmp"), ("All files", ".*")]
  if not is_macos: image_formats.insert(1, ("JPEG", (".jpg", ".jpeg")))


class Config:
  open = lambda path, mode='w': SDL_RWFromFile(path.encode("utf-8"), bytes(mode, "utf-8"))
  save_image = lambda surface, path: (SDL_SaveBMP_RW(surface, Config.open(path), 1) if path.endswith(".bmp") else
                                      IMG_SaveJPG_RW(surface, Config.open(path), 1, 70) if not Vars.is_macos and path.endswith((".jpg", ".jpeg")) else
                                      IMG_SavePNG_RW(surface, Config.open(path), 1))

  large_font = FontManager(Vars.path+"large_font.ttf", size=16)
  small_font = FontManager(Vars.path+"small_font.ttf", size=12)

  os_list = (                          #heap in byte, -1=infinite
    {"name": "PC",       "ratio": 0,   "heap": -1,     "color": "#4a4a4a", "unit": "deg",     "clock": False, "battery": "battery1.png"},
    {"name": "Numworks", "ratio": 1,   "heap": 65_536, "color": "#ffb531", "unit": "deg",     "clock": False, "battery": "battery0.png"},
    {"name": "Omega",    "ratio": 0.8, "heap": 99_000, "color": "#c53431", "unit": "sym/deg", "clock": True,  "battery": "battery0.png"},
    {"name": "Upsilon",  "ratio": 0.9, "heap": 69_500, "color": "#7ea2ce", "unit": "sym/deg", "clock": True,  "battery": "battery1.png"},
  )
  model_list = (
    {"name": "n0100", "ratio": 0, "disabled": True},
    {"name": "n0110", "ratio": 1},
    {"name": "n0120", "ratio": 0, "disabled": True},
  )

  default_os = 1
  default_model = 1
  zoom_max = 4

# Class to store some data
class StateData:
  def __init__(self, **content):
    self(**content)

  def __call__(self, **content):
    self.__dict__.update(content)
    return self

  def __str__(self):
    return "{}({})".format(__class__.__name__, ', '.join(f"{k}={repr(v)}" for k, v in self.__dict__.items() if not k.startswith("__")))

  __repr__ = __str__

  def __contains__(self, name):
    if type(name) != str: raise TypeError("attribute name must be an str")
    if not name: raise ValueError("empty attribute name")
    return name in self.__dict__



"""
Patch the module for MacOS. Redefine some methods of Core and Gui classes.

Since it's the only bone that doesn't accept the method I use to make a GUI,
it will try to patch the module to do a manual refresh.

This means that the program (if launched on this system) will have to regularly
call the display() method to refresh the window.

Basically using the same principle as in the first version of module
(when I was not using Threads yet)

Note: The window to notify the end of script to user, will be not available for mac
"""

from .core import Core, Gui, Vars, Draw, Colors

# Global instance
Core_self: Core = None

def Core___init__(self: Core):
    """Patched version of core, this will not start the Thread and just init library"""
    global Core_self # global instance for display function
    Core_self = self

    Gui.paused = Gui.already_paused = True

    # Stop module to avoid entering in main loop
    self.stopped = True
    # Init module
    self.event_loop()
    # And module is now available
    self.stopped = False

Core_event_fire_ = Core.event_fire
def Core_event_fire(self: Core, method, *args, **kwargs):
  """Wrapped method of .event_fire(), refresh the window after called the function"""
  value, err = Core_event_fire_(self, method, *args, **kwargs)
  if not self.stopped: Gui.refresh()
  return value, err

def Core_is_alive(self: Core):
  """Redefine Thread.is_alive() to return always True, because the Thread never started and doesn't want errors about this"""
  return True


# Set new methods in Core
Core.__init__ = Core___init__
Core.event_fire = Core_event_fire
Core.is_alive = Core_is_alive


# And define the display method
def display():
  """Refresh manually the window and display changes"""
  if Core_self is None:
    Gui.refresh() # unsafe
    return

  Core_self.refreshed = False
  _, err = Core_self.event_fire(Gui.refresh)
  Core_self.refreshed = True
  if err != None:
    raise err


# Also I don't know why, but tkinter give wrong widget id, so let's patch this
# to install: pyobjc-core pyobjc-framework-Cocoa
try: from objc import pyobjc_id
except ImportError as e:
  e.msg = "'pyobjc' module and his Cocoa framework is needed for your platform. Please install it with command 'pip3 install pyobjc-core pyobjc-framework-Cocoa'"
  raise

def Gui_get_widget_id(widget):
  """
  Returns the main window, not the widget frame because a can't get the widget with 'objc', there is nothing on web.
  Literally nobody has this problem, and no AI can solve my problem.

  So to avoid overlapping SDL windows, I'm going to offset them myself when refreshing
  """
  from AppKit import NSApp
  return pyobjc_id(NSApp.windows()[-1].contentView())

def Gui_refresh():
  """Offset drawable surface on screen"""
  Gui.created()

  Draw.rect(Gui.screen_surf, Colors.black)
  Draw.blit(Gui.screen_surf, Gui.head_surface)
  Draw.blit(Gui.screen_surf, Gui.drawable, (0, Vars.head_size))
  Gui.screen.refresh()
  Gui.tkmaster.update_idletasks()
  Gui.tkmaster.update()

# Set new methods in Gui
Gui.get_widget_id = Gui_get_widget_id
Gui.refresh = Gui_refresh



"""
A start of a 3D engine. Use``OK`` to display the faces or edges of cube and ``←↑→↓+-`` to move it.

**NOTE**: This script work better on Upsilon because ``fill_polygon()`` function doesn't exist in others OS. So we need to re-implement this and is consuming a lot of resources.
Source: https://my.numworks.com/python/antarctus/cube_move
"""
try:
  import os
  if hasattr(os, "environ"):
    os.environ["KANDINSKY_OS_MODE"]= '0'
    os.environ['KANDINSKY_ZOOM_RATIO'] = '2'
except: pass
from math import cos,sin,pi
from kandinsky import *
from time import sleep
from ion import *


MODEL=[]
COLORS=[(255,0,0),(255,255,0),(0,255,0),(0,255,255),(0,0,255),(255,0,255)]
FILL_MODEL=False

def cube3d(x1,y1,z1,x2,y2,z2):
  if FILL_MODEL:
    points3d([[x1,y1,z1],[x1,y2,z1],[x2,y2,z1],[x2,y1,z1]])
    points3d([[x1,y2,z1],[x1,y2,z2],[x2,y2,z2],[x2,y2,z1]])
    points3d([[x1,y1,z2],[x1,y2,z2],[x2,y2,z2],[x2,y1,z2]])
    points3d([[x1,y1,z2],[x1,y2,z2],[x1,y2,z1],[x1,y1,z1]])
    points3d([[x2,y1,z2],[x2,y1,z1],[x2,y2,z1],[x2,y2,z2]])
    points3d([[x2,y1,z2],[x2,y1,z1],[x1,y1,z1],[x1,y1,z2]])

  else:
    points3d([[x1,y1,z1],[x1,y2,z1]])
    points3d([[x1,y1,z1],[x2,y1,z1]])
    points3d([[x2,y1,z1],[x2,y2,z1]])
    points3d([[x1,y2,z1],[x2,y2,z1]])
    points3d([[x1,y1,z2],[x1,y2,z2]])
    points3d([[x1,y1,z2],[x2,y1,z2]])
    points3d([[x2,y1,z2],[x2,y2,z2]])
    points3d([[x1,y2,z2],[x2,y2,z2]])
    points3d([[x1,y1,z1],[x1,y1,z2]])
    points3d([[x2,y1,z1],[x2,y1,z2]])
    points3d([[x1,y2,z1],[x1,y2,z2]])
    points3d([[x2,y2,z1],[x2,y2,z2]])

def points3d(XYZpoints):
  MODEL.append(XYZpoints)

# Compatibility
def line(x1,y1,x2,y2,c):
  w=x2-x1
  h=y2-y1
  if abs(w)>=abs(h):
    d=h/w
    for i in range(0,w,(w>0)*2-1):
      set_pixel(x1+i,y1+int(d*i+0.5),c)
  else:
    d=w/h
    for i in range(0,h,(h>0)*2-1):
      set_pixel(x1+int(d*i+0.5),y1+i,c)

def polygone(p,c):
    p = list(p)
    p_len = len(p)

    for y in range(min([[0,222]]+p, key=lambda v: v[1])[1], max([[0,0]]+p, key=lambda v: v[1])[1]):
      switches = []
      last_point = p_len-1

      for i in range(p_len):
        point_y, last_point_y = p[i][1], p[last_point][1]
        if ((point_y < y and last_point_y >= y) or (last_point_y < y and point_y >= y)) and point_y != last_point_y:
          switches.append(round(p[i][0]+1*(y-point_y)/(last_point_y-point_y)*(p[last_point][0]-p[i][0])))
        last_point = i

      switches.sort()
      for x in range(0, len(switches), 2):
        if switches[x] >= 320*2: break
        if switches[x+1] > 0: fill_rect(switches[x], y, switches[x+1]-switches[x], 1, c)

try:
  line=draw_line
  string=draw_string
except: string=lambda t,x,y,c,bg,f: draw_string(t,x,y,c,bg)
try: polygone=fill_polygon
except: pass

# 3d to 2d render
def render(cosx,sinx,cosy,siny,cosz,sinz):
  fill_rect(0,0,320,200,"black")

  for i in range(len(MODEL)):
    polygon = []
    
    for p in MODEL[i]:
      #Rotations
      x,y,z=p
      y, z = xturn(y,z,cosx,sinx)
      x, z = yturn(x,z,cosy,siny)
      x, y = zturn(x,y,cosz,sinz)
      
      #Projections
      px,py=int(x*100/(z+300))+160, int(y*100/(z+300))+100
      
      if FILL_MODEL:
        # remove parts masked by another
        ...

      polygon.append([px,py])

      #debug
      string("%d,%d"%(px,py), px, py, "white", "black", 1)
    string("%f,%f,%f"%(cosx,cosy,cosz), 5, 0, "white", "black", 1)
    string("%f,%f,%f"%(sinx,siny,sinz), 5, 12, "white", "black", 1)
      
    if FILL_MODEL: polygone(polygon,COLORS[i%6])
    else:
      x1,y1=polygon[0]
      x2,y2=polygon[1]
      line(x1,y1,x2,y2,COLORS[i//2%6])

# ???
def xturn(y,z,cosx,sinx):
  return y*cosx-z*sinx,y*sinx+z*cosx

def yturn(x,z,cosy,siny):
  return z*siny+x*cosy,z*cosy-x*siny

def zturn(x,y,cosz,sinz):
  return x*cosz-y*sinz,x*sinz+y*cosz

cube3d(-100,-100,-100,100,100,100)

xrot=yrot=zrot=xoffset=yoffset=0

fill_rect(0,0,320,222,(0,)*3)
string("OK: fill on/off"+(' '*13)+"Arrows/+/-: move",5,207,"white","black",True)
while True:
  xrot+=pi/90*(keydown(KEY_UP)-keydown(KEY_DOWN))
  yrot+=pi/90*(keydown(KEY_LEFT)-keydown(KEY_RIGHT))
  zrot+=pi/90*(keydown(KEY_MINUS)-keydown(KEY_PLUS))

  if keydown(KEY_OK):
    FILL_MODEL = not FILL_MODEL
    MODEL.clear()
    cube3d(-100,-100,-100,100,100,100)
    sleep(0.2)

  render(cos(xrot),sin(xrot),cos(yrot),sin(yrot),cos(zrot),sin(zrot))
#  sleep(0.01)


"""A small module allowing to link the kandinsky module, from the Numworks, to a Windows window.
Useful if you want to test your program without putting it on the calculator.
The [GitHub project](https://github.com/ZetaMap/Kandinsky-Numworks), and if you have any questions, check out the [FAQ](https://github.com/ZetaMap/Kandinsky-Numworks/blob/main/FAQ.md)
In addition, this module also emulates the drawing speed, and has [many other features](https://github.com/ZetaMap/Kandinsky-Numworks/blob/main/README.md#additional-features).
"""

import sys # test is macos to patch the module
try:
  if sys.platform == 'darwin': from .util.mac_patcher import Core as __Core, display
  else: from .util.core import Core as __Core
except ImportError as e:
  if "relative import" not in e.msg:
    raise
  if sys.platform == 'darwin': from util.mac_patcher import Core as __Core, display
  else: from util.core import Core as __Core

__name__ = "kandinsky"
__version__ = "2.6.5"
try: 
  with open("README.md", encoding="utf-8") as f: __doc__ = f.read()
  del f
except (FileNotFoundError, OSError): __doc__ = "<unknown>"
__all__ = [
  #numworks
  "get_pixel",
  "set_pixel",
  "color",
  "draw_string",
  "fill_rect",

  #only for PC
  "quit",

  #omega
  "draw_line",
  "wait_vblank",
  "get_keys",
  "small_font",
  "large_font",

  #upsilon
  "draw_circle",
  "fill_circle",
  "fill_polygon",
  "get_palette"
]
if sys.platform == 'darwin': __all__.append("display")
__Core = __Core()

# No more need of this
del sys

# Numworks basic methods
def get_pixel(x, y, /):
  """Return pixel (x, y) color"""
  pixel, err = __Core.event_fire(__Core.get_pixel, x, y)
  if err != None:
    raise err
  return pixel

def set_pixel(x, y, color, /):
  """Color pixel (x, y)"""
  _, err = __Core.event_fire(__Core.set_pixel, x, y, color)
  if err != None:
    raise err

def color(r, g=None, b=None, /):
  """Define a rgb color"""
  color, err = __Core.event_fire(__Core.color,r, g, b)
  if err != None:
    raise err
  return color

# Argument added by Omega
if __Core.OS_MODE != 1:
  def draw_string(text, x, y, color=(0,0,0), background=(248,252,248), font=False, /):
    """Display a text from pixel (x, y)"""
    _, err = __Core.event_fire(__Core.draw_string, text, x, y, color, background, font)
    if err != None:
      raise err
else :
  def draw_string(text, x, y, color=(0,0,0), background=(248,252,248), /):
    """Display a text from pixel (x, y)"""
    _, err = __Core.event_fire(__Core.draw_string, text, x, y, color, background)
    if err != None:
      raise err

def fill_rect(x, y, width, height, color, /):
  """Fill a rectangle at pixel (x, y)"""
  _, err = __Core.event_fire(__Core.fill_rect, x, y, width, height, color)
  if err != None:
    raise err

# new method of kandinsky, only on PC
def quit():
  """Close manualy the window without notifying the user"""
  _, err = __Core.event_fire(__Core.quit_app)
  if err != None:
    raise err

# omega methods
def draw_line(x1, y1, x2, y2, color, /):
  """Draw a line at (x1, y1) to (x2, y2)"""
  _, err = __Core.event_fire(__Core.draw_line, x1, y1, x2, y2, color)
  if err != None:
    raise err

def wait_vblank():
  """Wait for screen refresh"""
  _, err = __Core.event_fire(__Core.wait_vblank)
  if err != None:
    raise err

def get_keys():
  """Get name of pressed keys"""
  keys, err = __Core.event_fire(__Core.get_keys)
  if err != None:
    raise err
  return keys

# some fields of omega, define size of font for python
small_font = False
large_font = True

# upsilon methods
def draw_circle(x, y, r, color, /):
  """Draw circle at (x, y) of radius r"""
  _, err = __Core.event_fire(__Core.draw_circle, x, y, r, color)
  if err != None:
    raise err

def fill_circle(x, y, r, color, /):
  """Fill circle at (x, y) of radius r"""
  _, err = __Core.event_fire(__Core.fill_circle, x, y, r, color)
  if err != None:
    raise err

def fill_polygon(points, color, /):
  """Fill polygon at points [(x1, y1), ...]"""
  _, err = __Core.event_fire(__Core.fill_polygon, points, color)
  if err != None:
    raise err

def get_palette():
  """Get theme palette"""
  pal, err = __Core.event_fire(__Core.get_palette)
  if err != None:
    raise err
  return pal


######### Clean #########
# delete methods according to OS
# in pc mode, do not delete anything
if __Core.OS_MODE:
  # upsilon moved this method and epsilon don't have this
  if __Core.OS_MODE != 2:
    __all__.remove("get_keys")
    del get_keys

  # delete upsilon methods
  if __Core.OS_MODE < 3:
    __all__.remove("draw_circle")
    __all__.remove("fill_circle")
    __all__.remove("fill_polygon")
    __all__.remove("get_palette")
    del draw_circle, fill_circle,  fill_polygon, get_palette

  # delete omega methods
  if __Core.OS_MODE < 2:
    __all__.remove("draw_line")
    __all__.remove("wait_vblank")
    __all__.remove("small_font")
    __all__.remove("large_font")
    del draw_line, wait_vblank, small_font, large_font



__all__ = ["Tests"]

def create_test(message, *valid_types):
  """
  Decorator to create the test. Put this before an empty method with (*objects, raise_err=True) arguments.

  The message can contains {found} and {valid} keys to format found type and required types.
  """
  if type(message) != str: raise TypeError("message must be an str")
  if len(valid_types) == 0: raise IndexError("one type or more must be specified")
  if not all(issubclass(type(t), type) for t in valid_types): raise TypeError("valid_types must be a list of types")

  def decorator(func):
    type_msg = valid_types[0].__name__ if len(valid_types) == 1 else (", ".join([i.__name__ for i in valid_types[:-1]])+" or "+valid_types[-1].__name__)
    def wrapper(*objects, raise_err=True):
      for i in objects:
        if type(i) not in valid_types:
          if raise_err: raise TypeError(message.format(found=type(i).__name__, valid=type_msg))
          else: return False

      return True
    return wrapper
  return decorator

class Tests:
  @create_test("can't convert {found} to {valid}", int)
  def int(*objects, raise_err=True): ...

  @create_test("can't convert {found!r} object to {valid} implicitly", str)
  def str(*objects, raise_err=True): ...

  @create_test("object {found!r} isn't a {valid}", tuple, list)
  def list(*objects, raise_err=True): ...



######### Imports #########
import sys, os

try: from tkinter import Tk
except ImportError as e:
  e.msg = "tkinter is not installed. "
  if sys.platform == "win32": e.msg += "please reinstall python with complements modules"
  elif sys.platform == "darwin": e.msg += "install it with 'brew install python-tk'"
  else: e.msg += "install it with 'sudo apt install python3-tk'"
  raise

import warnings
warnings.filters = [] # Reset filters because some default appear in, and HIS DON'T PRINT MY WARNINGS!!
warnings.simplefilter('ignore', UserWarning) # Disable SDL warning
try:
  import sdl2dll
  from sdl2.ext import init, quit as sdl_quit, ttf, image
  from sdl2._internal import prettywarn

  # I'm forced to not initialise the video for now,
  # due to MacOS causing a C++ crash if tkinter is initialized after SDL
  init(video=False)
  ttf._ttf_init()
  image._sdl_image_init()

except (ImportError, RuntimeError) as e:
  e.msg = """PySDL2 or PySDL2-dll module is not installed on your system, and kandinsky depend to it.
>>> To install, follow steps on the GitHub project: https://github.com/ZetaMap/Kandinsky-Numworks
>>> or type 'pip3 install -U pysdl2 pysdl2-dll"""
  raise

# import of ion moved
ION_PRESENT = False

from threading import Thread, main_thread
from math import sqrt
from .stuff import *


######### Environ config #########
DEBUG = 'KANDINSKY_ENABLE_DEBUG' in os.environ #or True

NO_GUI = 'KANDINSKY_NO_GUI' in os.environ

# '0': PC, '1': Numworks, '2': Omega, '3': Upsilon
mode = os.environ.get('KANDINSKY_OS_MODE')
if mode and mode.isdecimal(): Vars.selected_os = int(mode) if 0 <= int(mode) < len(Config.os_list) else Config.default_os

# '0': n0100, '1': n0110, '2': n0120
model = os.environ.get('KANDINSKY_MODEL_MODE')
if model and model.isdecimal(): Vars.selected_model = int(model) if 0 <= int(model) < len(Config.model_list) else Config.default_model

if 'KANDINSKY_SCREEN_SIZE' in os.environ:
  screen = os.environ['KANDINSKY_SCREEN_SIZE'].split('x')

  if len(screen) != 2:
    raise \
      ValueError("invalid screen format. format to use: '<width>x<height>'")

  for i in range(2):
    if not screen[i].isdecimal() or int(screen[i]) < 0:
      raise \
        ValueError("screen dimensions must be positive integers")
    screen[i] = int(screen[i])

  screen[1] -= Vars.head_size
  if screen[0] < Vars.screen[0] or screen[1] < Vars.screen[1]:
    raise \
      ValueError(f"screen dimensions must be greater than {Vars.screen[0]}x{Vars.screen[1]+Vars.head_size}")

  Vars.screen = tuple(screen)

zoom = os.environ.get('KANDINSKY_ZOOM_RATIO')
if zoom and zoom.isdecimal() and 1 <= int(zoom) <= Config.zoom_max: Vars.zoom_ratio = int(zoom)

# Experimental way to get more real emulation of numworks
import threading
USE_HEAP = 'KANDINSKY_USE_HEAP' in os.environ and hasattr(threading, "get_native_id")
# same problem than warning of ion
if USE_HEAP: prettywarn("python heap limitator is an experimental feature, so it can crach python several times", ImportWarning)

# We need to more the import of ion here, to not import it if os mode isn't Omega
if Vars.selected_os == 2:
  try: 
    with warnings.catch_warnings(record=True):
      import ion
  except ImportError: pass
  else:
    # Another module is called ion, so verify default function to see if it's the right module
    ION_PRESENT = hasattr(ion, "keydown")
    if not ION_PRESENT:
      prettywarn("ion is installed but no basic methods found. "
                 "Are you sure you have installed the correct module? "
                 "His pypi name is 'ion-numworks'.", ImportWarning)

    # In older version i forgot to define __version__ field
    # so if is not defined, is an outdated version.
    # Or verify if version is good
    elif not hasattr(ion, "__version__") or tuple([int(i) for i in ion.__version__.split('.') if i.isdecimal()]) < (2, 0):
      prettywarn("outdated version of 'ion-numworks', please update it", ImportWarning)
      ION_PRESENT = False


######### Cleanup #########
del mode, model, zoom, init, sys, os, sdl2dll, ttf, image, warnings, threading, #prettywarn


######### Main code #########
__all__ = ["Core"]

class Core(Thread):
  stopped = False
  asknoclose = False
  OS_MODE = Config.default_os

  def __init__(self):
    Gui.paused = Gui.already_paused = True # Pause events to give the time of thread to initialize

    super().__init__(None, self.event_loop, Vars.app_name.replace(' ', '')+self.__class__.__name__)
    self.start()
    while Gui.paused and self.is_alive(): usleep(100) # Wait a little to synchronize it

  def quit_app(self, *_):
    if not self.stopped:
      self.stopped = True
      Gui.paused = False # Now all calls of kandinsky raise an error

      usleep(10**4) # Wait a little to give time to draw methods finished
      Gui.destroy()
      sdl_quit()

  def print_debug(self, type, *messages, type_length=5, **args):
    if DEBUG:
      print("DEBUG: ", end='')
      print(type+' '*(type_length-len(type))+':', *messages, **args)

  def verify(self, method, *args, **kwargs):
    self.print_debug("Event", " "+method.__name__, (*args, *[f"{k}={repr(v)}" for k, v in kwargs.items()]), sep='')

    if Gui.paused: self.print_debug("Pause", "waiting...")
    while Gui.paused and self.is_alive(): usleep(1000)

    if self.stopped: raise RuntimeError("kandinsky window destroyed")
    elif not self.is_alive(): raise RuntimeError(f"an internal error has occurred. Stack trace is before this it.")

  def sleep(self, delay_us):
    if USE_HEAP and Gui.heap_set: return # No need to sleep

    ratio = Gui.data.model*Gui.data.ratio
    delay = int(delay_us*ratio)

    self.print_debug("Delay", "input =", delay_us, "µs, ratio ≃", str(round(ratio, 6))+", output =", delay, "µs")
    if ratio > 0: usleep(delay)

### methods
  def get_pixel(self, x, y):
    Tests.int(x, y)

    if x < 0 or x >= Vars.screen[0] or y < 0 or y >= Vars.screen[1]: color = Colors.black
    else: color = Colors.convert(Draw.get_at(Gui.drawable, x, y))

    self.sleep(77)
    return color

  def set_pixel(self, x, y, color):
    Tests.int(x, y)

    Draw.pixel(Gui.drawable, Colors.convert(color), x, y)
    self.sleep(96)

  def color(self, r, g, b):
    if g is None and b is None: color = Colors.convert(r)
    elif g is not None and b is not None:
      Tests.int(r, g, b)
      color = Colors.convert((r, g, b))
    else: raise TypeError("color takes 1 or 3 arguments")

    self.sleep(100)
    return color

  def draw_string(self, text, x, y, color, background, font=False):
    Tests.str(text)
    Tests.int(x, y)

    color = Colors.convert(color)
    background = Colors.convert(background)
    bs = (7, 14) if font else (10,18)
    font = Config.small_font if font else Config.large_font
    if '\0' in text: text = text[:text.index('\0')]

    for i, l in enumerate(text.replace('\r', '').replace('\f', '\x01').replace('\b', '\x01').replace('\v', '\x01').replace('\t', "    ").split('\n')):
      Draw.rect(Gui.drawable, background, (x*(not i), y+i*bs[1], len(l)*bs[0], bs[1]))
      Draw.string(Gui.drawable, font, l, x*(not i), y+i*bs[1]-2, color)
    self.sleep(86*(len(text)+(len(text)>=5)//1.2))

  def fill_rect(self, x, y, width, height, color):
    Tests.int(x, y, width, height)

    if width < 0:  x, width = x+width, -width
    if height < 0: y, height = y+height, -height

    Draw.rect(Gui.drawable, Colors.convert(color), (x, y, width, height))
    self.sleep(130+width*height/20+width*height*0.02)

  def wait_vblank(self):
    """why this?"""
    while not self.refreshed: usleep(100)

  def get_keys(self):
    """Don't tell me about this XD, it's omega"""
    # Raise an error if Ion-Numworks is not installed
    if not ION_PRESENT: raise NotImplementedError("please install or upgrade 'ion-numworks' before using this method")
    # Use the internal API because method doesn't exist with OS_MODE=2
    return getattr(ion, "__Ion").get_keys() # use getattr() because '__Ion' will be replaced by '_Core__Ion'

  def draw_line(self, x1, y1, x2, y2, color):
    """https://github.com/UpsilonNumworks/Upsilon/blob/upsilon-dev/kandinsky/src/context_line.cpp"""
    Tests.int(x1, y1, x2, y2)
    color = Colors.convert(color)

    s = abs(y2-y1) > abs(x2-x1)
    if s: x1, y1, x2, y2 = y1, x1, y2, x2
    if x1 > x2: x1, x2, y1, y2 = x2, x1, y2, y1
    g = 1 if x2 == x1 else (y2-y1)/(x2-x1)

    for x in range(x1, x2):
      ty = int(y1+g*(x-x1))
      Draw.pixel(Gui.drawable, color, *((ty, x) if s else (x, ty)))  

    self.sleep(1)

  def draw_circle(self, x, y, r, color):
    """https://github.com/UpsilonNumworks/Upsilon/blob/upsilon-dev/kandinsky/src/context_circle.cpp"""
    Tests.int(x, y, r)

    color = Colors.convert(color)
    xc, yc, m = 0, abs(r), 5-4*abs(r)

    while xc <= yc:
      Draw.pixel(Gui.drawable, color, xc+x, yc+y)
      Draw.pixel(Gui.drawable, color, yc+x, xc+y)
      Draw.pixel(Gui.drawable, color, -xc+x, yc+y)
      Draw.pixel(Gui.drawable, color, -yc+x, xc+y)
      Draw.pixel(Gui.drawable, color, xc+x, -yc+y)
      Draw.pixel(Gui.drawable, color, yc+x, -xc+y)
      Draw.pixel(Gui.drawable, color, -xc+x, -yc+y)
      Draw.pixel(Gui.drawable, color, -yc+x, -xc+y)

      if m > 0: yc, m = yc-1, m-8*yc
      xc, m = xc+1, m+8*xc+4

    self.sleep(sqrt(r)*(10*sqrt(10)*sqrt(r)-r))

  def fill_circle(self, x, y, r, color):
    """https://github.com/UpsilonNumworks/Upsilon/blob/upsilon-dev/kandinsky/src/context_circle.cpp"""
    Tests.int(x, y, r)
    color = Colors.convert(color)

    r = abs(r)
    for i in range(x-r, x+r):
      semi = int(sqrt(r**2-(x-i)**2))
      Draw.rect(Gui.drawable, color, (i, y-semi, 1, semi*2))

    self.sleep(sqrt(r)*(r+10*sqrt(10)*sqrt(r)-10))

  def fill_polygon(self, points, color):
    """https://github.com/UpsilonNumworks/Upsilon/blob/upsilon-dev/kandinsky/src/context_polygon.cpp"""
    Tests.list(points)
    if len(points) < 3: raise ValueError("polygon must have least 3 points")
    tests = [[len(i) for v in i if len(i) != 2 or not Tests.int(v)] for i in points if Tests.list(i)]
    if len(tests) and len(tests[0]): raise ValueError("requested length 2 but object has length "+str(tests[0][0]))

    color = Colors.convert(color)
    points = list(points)
    points_size = len(points)

    for y in range(min([[0,Vars.screen[0]]]+points, key=lambda v: v[1])[1], max([[0,0]]+points, key=lambda v: v[1])[1]):
      switches = []
      last_point = points_size-1

      for i in range(points_size):
        point_y, last_point_y = points[i][1], points[last_point][1]
        if ((point_y < y and last_point_y >= y) or (last_point_y < y and point_y >= y)) and point_y != last_point_y:
          switches.append(round(points[i][0]+1*(y-point_y)/(last_point_y-point_y)*(points[last_point][0]-points[i][0])))
        last_point = i

      switches.sort()
      for x in range(0, len(switches), 2):
        if switches[x] >= Vars.screen[1]*2: break
        if switches[x+1] > 0: Draw.rect(Gui.drawable, color, (switches[x], y, switches[x+1]-switches[x], 1))

    self.sleep(len(points))

  def get_palette():
    return {"Toolbar":        Gui.data.color,
            "AccentText":     (0, 132, 120),
            "HomeBackground": Colors.white,
            "PrimaryText":    Colors.black,
            "SecondaryText":  (104, 108, 104)}
###

  def event_loop(self):
    if USE_HEAP: Gui._main_thread_pid = self.native_id

    try:
      Gui(Tk())
      Gui.config(not NO_GUI)
    except RuntimeError as e:
      # Handle multiple import of library
      # Library cannot open new tkinter root windows while another is opened but it can open another if previous is destroyed
      # Note: no data from the previous window can be restored to new one
      if "main thread is not in main loop" in e.args[0]: prettywarn("multiple import of kandinsky is not permitted", ImportWarning)
      else: self.print_debug("ERROR", type(e).__name__+": "+e.args[0])
      return

    self.OS_MODE = Gui.os_mode.get()
    # Bind more event
    Gui.tkmaster.protocol("WM_DELETE_WINDOW", self.quit_app)
    Gui.tkmaster.bind("<Control-q>", self.quit_app)

    Gui.paused = Gui.already_paused = False # Initialization finished, unpause events

    while True:
      if self.stopped:
        self.quit_app()
        return
      elif not self.asknoclose and not main_thread().is_alive():
        if Gui.askscriptend(): self.stopped = True
        else: self.asknoclose = True

      self.refreshed = False
      try: Gui.refresh()
      except AttributeError: pass
      except RuntimeError: self.stopped = True
      usleep(1000)
      self.refreshed = True

  def event_fire(self, method, *arg, **kwargs):
    try:
      self.verify(method, *arg, **kwargs)
      return method(*arg, **kwargs), None
    except (Exception, BaseException) as e:
      return None, Exception.with_traceback(
        KeyboardInterrupt(type(e).__name__+": "+' '.join(e.args)) if isinstance(e, RuntimeError) else e,
        e.__traceback__.tb_next if DEBUG else None)



import sys, os
from setuptools import setup

# Used by setup script to restore the version to 'null' in __init__.py after builded the library
RESTORE_VERSION = "--version-null" in sys.argv

# Version of library
VERSION = "2.6.5"

# Get the absolute path
PATH = __file__[:__file__.rfind('\\')+1 or __file__.rfind('/')+1]

# Copy README and FAQ
with open(PATH+"src/kandinsky/README.md", "rt", encoding="utf-8") as new:
  DOC = new.read()
  with open(PATH+"README.md", "wt", encoding="utf-8") as old:
    old.write(DOC)

with open(PATH+"src/kandinsky/FAQ.md", "rt", encoding="utf-8") as new:
  with open(PATH+"FAQ.md", "wt", encoding="utf-8") as old:
    old.write(new.read())

# Set the version in __init__.py
with open(PATH+"src/kandinsky/__init__.py", "r+t", encoding="utf-8") as f:
  lines = f.readlines()
  for i in range(len(lines)):
    if "__version__" in lines[i]:
      if RESTORE_VERSION:
        if VERSION in lines[i]: lines[i] = lines[i].replace(VERSION, "null")
        else: lines[i] = f"__version__ = \"null\"\n"
      elif "null" in lines[i]: lines[i] = lines[i].replace("null", VERSION)  
      else: lines[i] = f"__version__ = \"{VERSION}\"\n"

      f.seek(0)
      f.writelines(lines)      
      break

# Stop here if restore version to 'null'
if RESTORE_VERSION: exit()

# Remove __pycache__ everywhere
def clean_pycache(path="."):
  """Clean __pycache__ directories recursively. Call this before setup()."""
  for file in os.listdir(path):
    new_path = os.path.join(path, file)
    if os.path.isdir(new_path): 
      if file == "__pycache__": 
        # Remove a file or a directory recursively using terminal commands.
        # This way avoid some permissions errors.
        if os.name == "nt": os.system(("rd /s" if os.path.isdir(new_path) else "del /f") + " /q \"" + new_path.replace('/', '\\') + "\"")
        else: os.system("rm -rf \"" + new_path.replace('\\', '/') + "\"")
      else: clean_pycache(new_path)
clean_pycache(PATH)

# Run setup
setup(
  name='kandinsky', 
  version=VERSION, 
  author='ZetaMap', 
  description='A small module allowing to link the kandinsky module, from the Numworks, to a window.', 
  license='MIT', 
  long_description=DOC, 
  long_description_content_type='text/markdown', 
  url='https://github.com/ZetaMap/Kandinsky-Numworks', 
  project_urls = {
    'GitHub Project': 'https://github.com/ZetaMap/Kandinsky-Numworks', 
    'My GitHub Page': 'https://github.com/ZetaMap/',
  }, 
  classifiers = [
    'Programming Language :: Python :: 3', 
    'License :: OSI Approved :: MIT License', 
    'Operating System :: Microsoft :: Windows', 
    'Operating System :: Unix', 
    'Operating System :: MacOS :: MacOS X',
  ], 
  package_dir = {"": "src"},
  packages = [
    "kandinsky",
    "kandinsky.util",
    "kandinsky.util.stuff",
    "kandinsky.util.data",
  ],
  package_data = {"": ["**"]},
  install_requires = [
    'pysdl2', 
    'pysdl2-dll',
    'pyobjc-core ; sys_platform=="darwin"',
    'pyobjc-framework-Cocoa ; sys_platform=="darwin"',
  ], 
  python_requires='>=3.6',
)



from typing import Tuple, List, Union, Optional, Dict, MutableSet
import sys

ColorInput = Union[Tuple[int, int, int], List[int, int, int], str, float]
ColorOutput = Tuple[int, int, int]

def get_pixel(x: int, y: int) -> ColorOutput: ...
def set_pixel(x: int, y: int, color: ColorInput) -> None: ...
def color(color: ColorInput) -> ColorOutput: ...
def color(r: int, g: int, b: int) -> ColorOutput: ...
def draw_string(text: str, x: int, y: int, color: Optional[ColorInput]=(0, 0, 0),
                background: Optional[ColorInput]=(248, 252, 248), font: bool=False) -> None: ...
def fill_rect(x: int, y: int, width: int, height: int, color: ColorInput) -> None: ...

# Special methods, only in pc
def quit() -> None: ...
if sys.platform == 'darwin': # Method only for macos
  def display() -> None: ...

# Methods for other os
def draw_line(x1: int, y1: int, x2: int, y2: int, color: ColorInput) -> None: ...
def wait_vblank() -> None: ...
def get_keys() -> MutableSet[str]: ...
small_font: bool = ...
large_font: bool = ...
def draw_circle(x: int, y: int, r: int, color: ColorInput) -> None: ...
def fill_circle(x: int, y: int, r: int, color: ColorInput) -> None: ...
def fill_polygon(points: List[Tuple[int, int]], color: ColorInput) -> None: ...
def get_palette() -> Dict[str, ColorOutput]: ...



# This script try to limit the heap size of python (max RAM to use).
#
# On Windows we don't have a dynamic heap limiter like Unix (Linux and MacOS)
# with the 'resource' module.
# So will manualy create classes and limit the actual job with his PID.
# This way call kernel methods so it can fail, if the throttling attempt fails,
# it will attempt to emulate latency with small latencies with 'time' module,
# but in general it works pretty well.
#
# For Linux and MacOS is more simple, will just import the 'resource' module
# and set the size python heap.
# And if the request fails, doing same thing of windows if failed.
#
#
# Windows example: https://stackoverflow.com/a/16791778
# Unix example: https://stackoverflow.com/a/53617847

from sys import platform
from sdl2._internal import prettywarn

__all__ = [
  "set_heap",
  "usleep"
]



if platform == "win32":
  import ctypes

  PROCESS_SET_QUOTA = 0x100
  PROCESS_TERMINATE = 0x1
  JobObjectExtendedLimitInformation = 9
  JOB_OBJECT_LIMIT_PROCESS_MEMORY = 0x100

  class IO_COUNTERS(ctypes.Structure):
    _fields_ = [
      ('ReadOperationCount', ctypes.c_uint64),
      ('WriteOperationCount', ctypes.c_uint64),
      ('OtherOperationCount', ctypes.c_uint64),
      ('ReadTransferCount', ctypes.c_uint64),
      ('WriteTransferCount', ctypes.c_uint64),
      ('OtherTransferCount', ctypes.c_uint64)
    ]

  class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [
      ('PerProcessUserTimeLimit', ctypes.c_int64),
      ('PerJobUserTimeLimit', ctypes.c_int64),
      ('LimitFlags', ctypes.c_uint32),
      ('MinimumWorkingSetSize', ctypes.c_void_p),
      ('MaximumWorkingSetSize', ctypes.c_void_p),
      ('ActiveProcessLimit', ctypes.c_uint32),
      ('Affinity', ctypes.c_void_p),
      ('PriorityClass', ctypes.c_uint32),
      ('SchedulingClass', ctypes.c_uint32)
    ]

  class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [
      ('BasicLimitInformation', JOBOBJECT_BASIC_LIMIT_INFORMATION),
      ('IoInfo', IO_COUNTERS),
      ('ProcessMemoryLimit', ctypes.c_void_p),
      ('JobMemoryLimit', ctypes.c_void_p),
      ('PeakProcessMemoryUsed', ctypes.c_void_p),
      ('PeakJobMemoryUsed', ctypes.c_void_p)
    ]

  # Set memory limit for process with specfied 'pid', to specified 'size' in bytes
  def set_heap(pid, size):
    assert size < 0 or size > 4095, "heap size must be greater than 4096 bytes or -1"

    job_info = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()

    job = ctypes.windll.kernel32.CreateJobObjectA(None, None)
    assert job != 0, "ouch... that hurt, fatal error"

    assert ctypes.windll.kernel32.QueryInformationJobObject(
      job,
      JobObjectExtendedLimitInformation,
      ctypes.POINTER(JOBOBJECT_EXTENDED_LIMIT_INFORMATION)(job_info),
      ctypes.sizeof(JOBOBJECT_EXTENDED_LIMIT_INFORMATION),
      ctypes.POINTER(ctypes.c_uint32)(ctypes.c_uint32())
    ), "cannot create job handle"

    job_info.BasicLimitInformation.LimitFlags |= JOB_OBJECT_LIMIT_PROCESS_MEMORY
    job_info.ProcessMemoryLimit = size
    assert ctypes.windll.kernel32.SetInformationJobObject(
      job,
      JobObjectExtendedLimitInformation,
      ctypes.POINTER(JOBOBJECT_EXTENDED_LIMIT_INFORMATION)(job_info),
      ctypes.sizeof(JOBOBJECT_EXTENDED_LIMIT_INFORMATION)
    ), "cannot create new job informations, probably invalid heap size"

    process = ctypes.windll.kernel32.OpenProcess(PROCESS_SET_QUOTA | PROCESS_TERMINATE, False, pid)
    assert process != 0, f"process '{pid}' not found"

    assert ctypes.windll.kernel32.AssignProcessToJobObject(job, process), "cannot assign new process informations"
    assert ctypes.windll.kernel32.CloseHandle(job), "failed to closing job handle"
    assert ctypes.windll.kernel32.CloseHandle(process), "failed closing process handle"

else:
  import resource
  set_heap = lambda pid, size: resource.prlimit(pid, resource.RLIMIT_DATA, (size, resource.getrlimit(resource.RLIMIT_DATA)[1]))


# Try to find an real time clock to do microseconds sleeps
if platform == "win32":
  # Try to use module win-precise-time to get more sleeps precision
  try:
    import win_precise_time as _wpt
    usleep = lambda us: _wpt.sleep_until_ns(int(_wpt.time_ns()+us*1000))
  except:
    import time as _time
    prettywarn("win-precise-time module is not installed. Using time module to do usleep (emulation will be less accurate)", RuntimeWarning)
    def usleep(us):
      t = _time.perf_counter_ns()+us*1000
      while _time.perf_counter_ns() < t: _time.sleep(1e-9)

elif platform == "linux":
  try:
    import ctypes
    usleep = ctypes.CDLL("libc.so.6").usleep
    del ctypes
  except:
    # On some linux distribution, this library is not installed by default
    import time as _time
    prettywarn("libc6 library is not installed. Using time module to do usleep (emulation will be less accurate)", RuntimeWarning)
    usleep = lambda us: _time.sleep(us/10e6)

else:
  # MacOS don't have libc6 library and idk for other OS so just do a normal sleep
  import time as _time
  usleep = lambda us: _time.sleep(us/10e6)



# Other stuff, 
# because there is a bug when window changing monitor 
# (only on macos... again... this is a poor platform)
def get_monitors():
  import ctypes
  class RECT(ctypes.Structure):
    _fields_ = [
      ('left', ctypes.c_long),
      ('top', ctypes.c_long),
      ('right', ctypes.c_long),
      ('bottom', ctypes.c_long)
      ]
    def dump(self):
      return [int(val) for val in (self.left, self.top, self.right, self.bottom)]

  retval = {}
  def cb(hMonitor, hdcMonitor, lprcMonitor, dwData):
    retval.update({hMonitor: lprcMonitor.contents.dump()})
    return True
  ctypes.windll.user32.EnumDisplayMonitors(0, 0, ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(RECT), ctypes.c_double)(cb), 0)
  return retval
